#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABC, abstractmethod
from typing import Iterable, Mapping, Type, TypeVar, cast

from overrides import overrides

from ..request.conversation_request import ConversationRequest
from ..tweet.conversation_tweet_stream import ConversationTweetStream
from ..tweet.tweet import TweetId
from .retriever import Retriever, RetrieverBatch, RetrieverTweetStream


class ConversationRetrieverTweetStream(RetrieverTweetStream, ConversationTweetStream):
    # TODO: Access to tombstones count and meta-less Tweet-IDs
    pass


class ConversationRetrieverBatch(RetrieverBatch, ABC):
    def __init__(self, json: Mapping[str, Mapping[str, object]]):
        # TODO: See if this is necessary for ThreadRetriever
        self.num_tombstones = 0
        super().__init__(json)

    @abstractmethod
    @overrides
    def _tweet_ids(self) -> Iterable[TweetId]:
        pass


_T_ConversationRequest = TypeVar("_T_ConversationRequest", bound=ConversationRequest)


class ConversationRetriever(Retriever[_T_ConversationRequest], ABC):
    @classmethod
    @overrides
    def _tweet_stream_type(cls) -> Type[ConversationRetrieverTweetStream]:
        return ConversationRetrieverTweetStream

    @classmethod
    @overrides
    def _retriever_batch_type(cls) -> Type[ConversationRetrieverBatch]:
        return ConversationRetrieverBatch

    @property  # type: ignore  # see https://github.com/python/mypy/issues/1362
    @overrides
    def tweet_stream(self) -> "ConversationRetrieverTweetStream":
        return cast(ConversationRetrieverTweetStream, self._tweet_stream)

    @overrides
    def _timeline_url(self) -> Mapping[str, object]:
        return {"url": "https://mobile.twitter.com/_/status/" + self._request.tweet_id}

    @overrides
    def _batch_url(self) -> Mapping[str, object]:
        return {
            "url": (
                "https://api.twitter.com/2/timeline/conversation/{:s}.json".format(
                    self._request.tweet_id
                )
            ),
            "params": {
                "include_profile_interstitial_type": 1,
                "include_blocking": 1,
                "include_blocked_by": 1,
                "include_followed_by": 1,
                "include_want_retweets": 1,
                "include_mute_edge": 1,
                "include_can_dm": 1,
                "include_can_media_tag": 1,
                "skip_status": 1,
                "cards_platform": "Web-12",
                "include_cards": 1,
                "include_composer_source": "true",
                "include_ext_alt_text": "true",
                "include_reply_count": 1,
                "tweet_mode": "extended",
                "include_entities": "true",
                "include_user_entities": "true",
                "include_ext_media_color": "true",
                "include_ext_media_availability": "true",
                "send_error_codes": "true",
                "count": self._request.batch_size,
                "cursor": self._cursor,
                "ext": "mediaStats,highlightedLabel,cameraMoment",
            },
        }



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import re
from datetime import date, datetime, timedelta, timezone
from typing import Any, Tuple, cast

import pytest

from nasty.request.search import Search, SearchFilter


@pytest.mark.parametrize("max_tweets", [1, 10, 100, 1000], ids=repr)
def test_max_tweets(max_tweets: int) -> None:
    # Using batch_size=100 to speed up these larger requests and since we don't care
    # about accuracy to query here.
    tweets = list(
        Search(
            "trump", filter_=SearchFilter.LATEST, max_tweets=max_tweets, batch_size=100
        ).request()
    )
    assert max_tweets == len(tweets)
    assert len(tweets) == len({tweet.id for tweet in tweets})


# -- test_query_word_* -----------------------------------------------------------------


@pytest.mark.parametrize("word", ["trump", "hillary", "obama"], ids=repr)
def test_query_word_single(word: str) -> None:
    tweets = list(Search(word, filter_=SearchFilter.LATEST, max_tweets=50).request())
    assert 50 == len(tweets)
    for tweet in tweets:
        assert word.lower() in json.dumps(tweet.to_json()).lower()


# Random string that currently does not match any Tweet.
@pytest.mark.parametrize("word", ["c9dde8b5451149e683d4f07e4c4348ef"], ids=repr)
def test_query_word_unkown(word: str) -> None:
    assert not list(Search(word).request())


TEST_QUERY_KEYWORDS = [("trump", "hillary"), ("trump", "obama"), ("obama", "hillary")]


@pytest.mark.parametrize("args", TEST_QUERY_KEYWORDS, ids=repr)
def test_query_word_and(args: Tuple[str, str]) -> None:
    word1, word2 = args
    tweets = list(
        Search(
            "{} and {}".format(word1, word2), filter_=SearchFilter.LATEST, max_tweets=50
        ).request()
    )
    assert 50 == len(tweets)
    for tweet in tweets:
        all_tweet_text = json.dumps(tweet.to_json()).lower()
        assert word1.lower() in all_tweet_text
        assert word2.lower() in all_tweet_text


@pytest.mark.parametrize("args", TEST_QUERY_KEYWORDS, ids=repr)
def test_query_word_or(args: Tuple[str, str]) -> None:
    word1, word2 = args
    tweets = list(
        Search(
            "{} or {}".format(word1, word2), filter_=SearchFilter.LATEST, max_tweets=50
        ).request()
    )
    assert 50 == len(tweets)
    for tweet in tweets:
        all_tweet_text = json.dumps(tweet.to_json()).lower()
        assert word1.lower() in all_tweet_text or word2.lower() in all_tweet_text


@pytest.mark.parametrize("args", TEST_QUERY_KEYWORDS, ids=repr)
def test_query_word_not(args: Tuple[str, str]) -> None:
    word1, word2 = args
    tweets = list(
        Search(
            "{} -{}".format(word1, word2), max_tweets=50, filter_=SearchFilter.LATEST
        ).request()
    )
    assert 50 == len(tweets)
    for tweet in tweets:
        all_tweet_text = json.dumps(tweet.to_json()).lower()
        # Sadly, word2 can sometimes still occur in the Text even though we specifically
        # ask Twitter not to. In those cases I do not want to count this case a failure
        # and skip it then.
        assert word1.lower() in all_tweet_text
        if word2.lower() in tweet.text.lower():
            pytest.skip(
                "Negative query word '{}' found in result tweet: {}".format(
                    word2, tweet.to_json()
                )
            )


@pytest.mark.parametrize(
    "phrase", ["donald trump", "hillary clinton", "barack obama"], ids=repr
)
def test_query_word_phrase(phrase: str) -> None:
    tweets = list(Search(phrase, max_tweets=50, filter_=SearchFilter.LATEST).request())
    assert 50 == len(tweets)
    for tweet in tweets:
        all_tweet_text = json.dumps(tweet.to_json()).lower()
        if phrase.lower() in all_tweet_text:
            # Remove non alphanumeric, see https://stackoverflow.com/a/1277047/211404
            all_tweet_text = re.sub(r"[\W_]+", "", all_tweet_text)
            phrase = re.sub(r"[\W_]+", "", phrase.lower())
            assert phrase in all_tweet_text


# -- test_quest_user_* -----------------------------------------------------------------


TEST_QUERY_USERS = ["realDonaldTrump", "HillaryClinton", "BarackObama"]


@pytest.mark.parametrize("user", TEST_QUERY_USERS, ids=repr)
def test_query_user_from(user: str) -> None:
    tweets = list(
        Search("from:@" + user, max_tweets=50, filter_=SearchFilter.LATEST).request()
    )
    assert 0 < len(tweets) <= 50
    for tweet in tweets:
        assert user.lower() == tweet.user.screen_name.lower()


@pytest.mark.parametrize("user", TEST_QUERY_USERS, ids=repr)
def test_query_user_to(user: str) -> None:
    tweets = list(
        Search("to:@" + user, max_tweets=50, filter_=SearchFilter.LATEST).request()
    )
    assert 50 == len(tweets)
    for tweet in tweets:
        if not tweet.text.lower().count("@" + user.lower()):
            # Sometimes when a user creates a thread his individual Tweets will not
            # reply to the user, for example:
            # https://twitter.com/_/status/1197499643086753793
            assert user.lower() == tweet.user.screen_name.lower()


# -- test_date_* -----------------------------------------------------------------------


@pytest.mark.parametrize(
    "args",
    [
        # range: 1 year
        (date(2010, 1, 1), date(2010, 12, 31)),
        (date(2015, 1, 1), date(2015, 12, 31)),
        (date(2019, 1, 1), date(2019, 12, 31)),
        # range: 1 day
        (date(2010, 1, 1), date(2010, 1, 2)),
        (date(2015, 2, 10), date(2015, 2, 11)),
        (date(2019, 3, 21), date(2019, 3, 22)),
        # today
        (date.today() - timedelta(days=1), date.today() + timedelta(days=1)),
    ],
    ids=repr,
)
def test_date_range(args: Tuple[date, date]) -> None:
    since, until = args
    tweets = list(
        Search(
            "trump",
            since=since,
            until=until,
            filter_=SearchFilter.LATEST,
            max_tweets=40,
        ).request()
    )
    assert 40 == len(tweets)
    for tweet in tweets:
        assert since <= tweet.created_at.date() < until


def test_date_future() -> None:
    assert not list(Search("trump", since=(date.today() + timedelta(days=7))).request())


# -- test_filter_* ---------------------------------------------------------------------


def test_filter_top() -> None:
    assert 50 == len(
        list(Search("trump", filter_=SearchFilter.TOP, max_tweets=50).request())
    )
    # Since it is unknown how Twitter determines "top" tweets there is no way to check
    # for that.


def test_filter_latest() -> None:
    # Check if the 50 latest Tweets about "trump" are from the last 24h. Assumes that
    # each day there are at least 50 Tweets about "trump".
    tweets = list(Search("trump", filter_=SearchFilter.LATEST, max_tweets=50).request())
    assert 50 == len(tweets)
    for tweet in tweets:
        assert datetime.now(timezone.utc) - timedelta(days=1) < tweet.created_at


def test_filter_photos() -> None:
    tweets = list(Search("trump", filter_=SearchFilter.PHOTOS, max_tweets=50).request())
    assert 50 == len(tweets)
    for tweet in tweets:
        json = cast(Any, tweet.json)
        assert len(json["extended_entities"]["media"])
        for medium in json["extended_entities"]["media"]:
            assert medium["type"] in {"photo", "animated_gif"}


def test_filter_videos() -> None:
    tweets = list(Search("trump", filter_=SearchFilter.VIDEOS, max_tweets=50).request())
    assert 50 == len(tweets)
    for tweet in tweets:
        if "extended_entities" in tweet.json:
            # Video hosted on Twitter.
            json = cast(Any, tweet.json)
            assert len(json["extended_entities"]["media"])
            if "video" in tweet.text.lower():
                # Had one case, where an image post containing the substring
                # "VIDEO: youtu.be/..." matched this query.
                return
            for medium in json["extended_entities"]["media"]:
                assert "video" == medium["type"]
        else:
            # Video hosted  on external platform. AFAIK there is no general way to check
            # whether an URL to an external platform contains a video.
            pass


# -- test_lang_* -----------------------------------------------------------------------


def test_lang_en() -> None:
    assert 50 == len(
        list(
            Search(
                "trump", lang="en", max_tweets=50, filter_=SearchFilter.LATEST
            ).request()
        )
    )
    # No robust way to verify language.


def test_lang_de() -> None:
    assert 50 == len(
        list(
            Search(
                "trump", lang="de", max_tweets=50, filter_=SearchFilter.LATEST
            ).request()
        )
    )
    # No robust way to verify language.


def test_lang_invalid() -> None:
    assert not list(
        Search(
            "trump", lang="INVALID", max_tweets=50, filter_=SearchFilter.LATEST
        ).request()
    )


# -- test_special_msg_* ----------------------------------------------------------------


def test_special_msg_coronavirus() -> None:
    tweets = list(
        Search("coronavirus", max_tweets=50, filter_=SearchFilter.LATEST).request()
    )
    assert 50 == len(tweets)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import Counter
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from logging import getLogger
from os import getenv
from pathlib import Path
from tempfile import mkdtemp
from typing import Iterator, List, Optional, Sequence, Union, overload
from uuid import uuid4

from .._util.json_ import (
    JsonSerializedException,
    read_json,
    read_json_lines,
    write_json,
    write_jsonl_lines,
)
from ..request.request import Request
from ._execute_result import _ExecuteResult
from .batch_entry import BatchEntry
from .batch_results import BatchResults

logger = getLogger(__name__)


class Batch:
    def __init__(self) -> None:
        self._entries: List[BatchEntry] = []

    def append(self, request: Request) -> None:
        self._entries.append(
            BatchEntry(request, id_=uuid4().hex, completed_at=None, exception=None)
        )

    def dump(self, file: Path) -> None:
        logger.debug("Dumping batch to file '{}'.".format(file))
        write_jsonl_lines(file, self._entries, overwrite_existing=True)

    def load(self, file: Path) -> None:
        logger.debug("Loading batch from file '{}'.".format(file))
        self._entries += read_json_lines(file, BatchEntry)

    def execute(self, results_dir: Optional[Path] = None) -> Optional[BatchResults]:
        logger.debug(
            "Started executing batch of {:d} requests.".format(len(self._entries))
        )

        if not results_dir:
            results_dir = Path(mkdtemp())
        logger.debug("  Saving results to '{}'.".format(results_dir))
        Path.mkdir(results_dir, exist_ok=True, parents=True)

        num_workers = int(getenv("NASTY_NUM_WORKERS", default="1"))
        with ThreadPoolExecutor(max_workers=num_workers) as pool:
            futures = (
                pool.submit(self._execute_entry, entry, results_dir)
                for entry in self._entries
            )
            result_counter = Counter(
                future.result() for future in as_completed(futures)
            )

        logger.info(
            "Executing batch completed. "
            "{:d} successful, {:d} skipped, {:d} failed.".format(
                result_counter[_ExecuteResult.SUCCESS],
                result_counter[_ExecuteResult.SKIP],
                result_counter[_ExecuteResult.FAIL],
            )
        )
        if result_counter[_ExecuteResult.FAIL]:
            logger.error("Some requests failed!")
            return None
        return BatchResults(results_dir)

    @classmethod
    def _execute_entry(cls, entry: BatchEntry, results_dir: Path) -> _ExecuteResult:
        logger.debug("Executing request: {}".format(entry.request.to_json()))

        meta_file = results_dir / entry.meta_file_name
        data_file = results_dir / entry.data_file_name

        if meta_file.exists():
            prev_execution_entry = read_json(meta_file, BatchEntry)

            if data_file.exists():
                logger.debug("  Skipping request, because files already exist.")
                entry.completed_at = prev_execution_entry.completed_at
                return _ExecuteResult.SKIP

            logger.debug(
                "  Retrying request that has stray files from previous "
                "execution: {}".format(prev_execution_entry.exception)
            )
            meta_file.unlink()

        result = _ExecuteResult.SUCCESS
        try:
            write_jsonl_lines(data_file, entry.request.request(), use_lzma=True)
            entry.completed_at = datetime.now()
        except Exception as e:
            logger.exception("  Request execution failed with exception.")
            entry.exception = JsonSerializedException.from_exception(e)
            result = _ExecuteResult.FAIL

        write_json(meta_file, entry)
        return result

    def __len__(self) -> int:
        return len(self._entries)

    def __contains__(self, item: object) -> bool:
        return item in self._entries

    @overload
    def __getitem__(self, _index: int) -> BatchEntry:
        ...

    @overload  # noqa: F811
    def __getitem__(self, _slice: slice) -> Sequence[BatchEntry]:
        ...

    def __getitem__(  # noqa: F811
        self, index_or_slice: Union[int, slice]
    ) -> Union[BatchEntry, Sequence[BatchEntry]]:
        return self._entries[index_or_slice]

    def __iter__(self) -> Iterator[BatchEntry]:
        return iter(self._entries)

    def __repr__(self) -> str:
        return repr(self._entries)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from logging import getLogger
from pathlib import Path
from typing import Optional

from nasty_utils import (
    ColoredBraceStyleAdapter,
    LoggingSettings,
    Settings,
    SettingsConfig,
)
from pydantic import SecretStr, validator

_LOGGER = ColoredBraceStyleAdapter(getLogger(__name__))

_T_VALIDATOR = classmethod


def _key_validator(v: object) -> Optional[str]:
    if not isinstance(v, str) or v.startswith("Enter Twitter "):
        return None
    return v


class TwitterApiSettings(Settings):
    consumer_api_key: Optional[SecretStr]
    _consumer_api_key_validator: _T_VALIDATOR = validator(
        "consumer_api_key", pre=True, allow_reuse=True
    )(_key_validator)

    consumer_api_secret: Optional[SecretStr]
    _consumer_api_secret_validator: _T_VALIDATOR = validator(
        "consumer_api_secret", pre=True, allow_reuse=True
    )(_key_validator)

    access_token: Optional[SecretStr]
    _access_token_validator: _T_VALIDATOR = validator(
        "access_token", pre=True, allow_reuse=True
    )(_key_validator)

    access_token_secret: Optional[SecretStr]
    _access_token_secret_validator: _T_VALIDATOR = validator(
        "access_token_secret", pre=True, allow_reuse=True
    )(_key_validator)


class NastySettings(LoggingSettings):
    class Config(SettingsConfig):
        search_path = Path("nasty.toml")

    twitter_api: TwitterApiSettings



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from http import HTTPStatus

from typing_extensions import Final


class UnexpectedStatusCodeException(Exception):
    def __init__(
        self,
        url: str,
        status_code: HTTPStatus,
        *,
        expected_status_code: HTTPStatus = HTTPStatus.OK,
    ):
        self.url: Final = url
        self.status_code: Final = status_code
        self.expected_status_code: Final = expected_status_code

        super().__init__(
            "Received status code {:d} {:s} (expected {:d} {:s}) for URL "
            '"{:s}".'.format(
                self.status_code.value,
                self.status_code.name,
                self.expected_status_code.value,
                self.expected_status_code.name,
                self.url,
            )
        )



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from logging import getLogger
from time import sleep
from typing import Iterable, Mapping, Optional, cast

import tweepy
from more_itertools import chunked

from nasty._settings import NastySettings, TwitterApiSettings

from ..tweet.tweet import Tweet, TweetId

logger = getLogger(__name__)


def _make_tweepy_api(settings: TwitterApiSettings) -> tweepy.API:
    if settings.consumer_api_key is None or settings.consumer_api_secret is None:
        raise ValueError(
            "To use this you need to create a configuration file with your Twitter API "
            "keys. Copy file config-example.nasty.toml in the nasty source folder "
            "to ${{XDG_CONFIG_HOME}}/{} and fill out the respective values.".format(
                NastySettings.Config.search_path
            )
        )

    tweepy_auth: tweepy.auth.AuthHandler
    if settings.access_token is not None and settings.access_token_secret is not None:
        tweepy_auth = tweepy.OAuthHandler(
            settings.consumer_api_key.get_secret_value(),
            settings.consumer_api_secret.get_secret_value(),
        )
        tweepy_auth.set_access_token(
            settings.access_token.get_secret_value(),
            settings.access_token_secret.get_secret_value(),
        )
    else:
        tweepy_auth = tweepy.AppAuthHandler(
            settings.consumer_api_key.get_secret_value(),
            settings.consumer_api_secret.get_secret_value(),
        )
    return tweepy.API(tweepy_auth, parser=tweepy.parsers.JSONParser())


STATUSES_LOOKUP_CHUNK_SIZE = 100
CHUNK_LOOKUPS_SINCE_LAST_RATE_LIMIT_ERROR = 0
TWEEPY_API: Optional[tweepy.API] = None


def statuses_lookup(
    tweet_ids: Iterable[TweetId], twitter_api_settings: TwitterApiSettings
) -> Iterable[Optional[Tweet]]:
    global CHUNK_LOOKUPS_SINCE_LAST_RATE_LIMIT_ERROR
    global TWEEPY_API

    if TWEEPY_API is None:
        TWEEPY_API = _make_tweepy_api(twitter_api_settings)

    for tweet_ids_chunk in chunked(tweet_ids, STATUSES_LOOKUP_CHUNK_SIZE):
        num_retries = 0
        while True:
            exception: Exception
            try:
                tweets_chunk = cast(
                    Mapping[str, Mapping[str, Optional[Mapping[str, object]]]],
                    TWEEPY_API.statuses_lookup(
                        tweet_ids_chunk,
                        include_entities=True,
                        map_=True,
                        tweet_mode="extended",
                    ),
                )["id"]
                CHUNK_LOOKUPS_SINCE_LAST_RATE_LIMIT_ERROR += 1
                for tweet_id in tweet_ids_chunk:
                    tweet_json = tweets_chunk[tweet_id]
                    if tweet_json is not None:
                        yield Tweet(tweet_json)
                    else:
                        yield None
                break

            except tweepy.RateLimitError as e:
                logger.debug(
                    "Chunk lookups since last rate limit error: {}.".format(
                        CHUNK_LOOKUPS_SINCE_LAST_RATE_LIMIT_ERROR
                    )
                )
                CHUNK_LOOKUPS_SINCE_LAST_RATE_LIMIT_ERROR = 0

                logger.info("Hit rate limit error. Sleeping for 15mins...")
                sleep(15 * 60)
                logger.info("Retrying...")
                exception = e

            except Exception as e:
                logger.exception("Exception occurred.")
                logger.info("Retrying (retry {})...".format(num_retries))
                exception = e

            num_retries += 1
            if num_retries == 3:
                logger.error("Maximum number of retries exceeded.")
                raise exception



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nox import options, session
from nox.sessions import Session

options.error_on_external_run = True
options.reuse_existing_virtualenvs = True
options.stop_on_first_error = True


@session(python=["3.6", "3.7", "3.8", "pypy3"])
def test(session: Session) -> None:
    session.install("-e", ".[test]")
    session.run(
        "pytest",
        "--cov",
        "--cov-report=",
        "--cov-context",
        "test",
        "--html",
        "tests-report.html",
        "--self-contained-html",
    )



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import date, timedelta
from enum import Enum
from typing import Dict, Mapping, Optional, Sequence, cast

from overrides import overrides
from typing_extensions import Final

from .._util.time_ import daterange, yyyy_mm_dd_date
from .._util.typing_ import checked_cast
from ..tweet.tweet_stream import TweetStream
from .request import DEFAULT_BATCH_SIZE, DEFAULT_MAX_TWEETS, Request


class SearchFilter(Enum):
    """Different sorting/filtering rules for Twitter search results.

    - TOP: Sort result Tweets by popularity (e.g., when a lot of people are interacting
        with or sharing via Retweets and replies)
    - LATEST: Sort result Tweets by most-recent post date.
    - PHOTOS: To only see Tweets that includes photos.
    - VIDEOS: To only see Tweets that includes videos.

    See: https://help.twitter.com/en/using-twitter/top-search-results-faqs
    """

    TOP = "TOP"
    LATEST = "LATEST"
    PHOTOS = "PHOTOS"
    VIDEOS = "VIDEOS"

    def to_json(self) -> str:
        return self.name

    @classmethod
    def from_json(cls, obj: str) -> "SearchFilter":
        return cls[obj]


DEFAULT_FILTER = SearchFilter.TOP
DEFAULT_LANG = "en"


class Search(Request):
    def __init__(
        self,
        query: str,
        *,
        since: Optional[date] = None,
        until: Optional[date] = None,
        filter_: SearchFilter = DEFAULT_FILTER,
        lang: str = DEFAULT_LANG,
        max_tweets: Optional[int] = DEFAULT_MAX_TWEETS,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ):
        """Construct a new query.

        :param query: String that should be searched for. Twitter allows some advanced
            operations here, like exact phrase match, negative match, AND/OR, and
            to/from specific users. For more details, see:
            https://help.twitter.com/en/using-twitter/twitter-advanced-search

            There is no guarantee that the query string will be contained in the Tweet
            text. It could also be part of the name of the authoring user, or even the
            title of a linked external website.
        :param since: Only find Tweets written after this date (inclusive).
        :param until: Only find Tweets written before this date (exclusive).
        :param filter_: Method to sort/filter Tweets.
        :param lang: Only search Tweets written in this language. These are directly
            passed to Twitter and it's undocumented what arguments they except here.
            Presumably ISO 3166-1 alpha-2 and alpha-3 codes should work.
        """

        if since is not None and until is not None and since >= until:
            raise ValueError("since date must be before until date.")

        super().__init__(max_tweets=max_tweets, batch_size=batch_size)
        self.query: Final = query
        self.since: Final = since
        self.until: Final = until
        self.filter: Final = filter_
        self.lang: Final = lang

    @overrides
    def to_json(self) -> Mapping[str, object]:
        obj: Dict[str, object] = {
            "type": None,  # Will be set in super(), but forces order.
            "query": self.query,
        }
        if self.since:
            obj["since"] = self.since.isoformat()
        if self.until:
            obj["until"] = self.until.isoformat()
        obj["filter"] = self.filter.to_json()
        obj["lang"] = self.lang
        obj.update(super().to_json())
        return obj

    @classmethod
    @overrides
    def from_json(cls, obj: Mapping[str, object]) -> "Search":
        assert obj["type"] == cls.__name__
        return cls(
            query=checked_cast(str, obj["query"]),
            since=(
                yyyy_mm_dd_date(checked_cast(str, obj["since"]))
                if "since" in obj
                else None
            ),
            until=(
                yyyy_mm_dd_date(checked_cast(str, obj["until"]))
                if "until" in obj
                else None
            ),
            filter_=SearchFilter.from_json(cast(str, obj["filter"])),
            lang=checked_cast(str, obj["lang"]),
            max_tweets=(
                cast(Optional[int], obj["max_tweets"])
                if "max_tweets" in obj
                else DEFAULT_MAX_TWEETS
            ),
            batch_size=(
                checked_cast(int, obj["batch_size"])
                if "batch_size" in obj
                else DEFAULT_BATCH_SIZE
            ),
        )

    @overrides
    def request(self) -> TweetStream:
        from .._retriever.search_retriever import SearchRetriever

        return SearchRetriever(self).tweet_stream

    def to_daily_requests(self) -> Sequence["Search"]:
        if self.since is None or self.until is None:
            raise ValueError(
                "Need both since and until date for into_daily_requests()."
            )

        return [
            Search(
                self.query,
                since=date_,
                until=date_ + timedelta(days=1),
                filter_=self.filter,
                lang=self.lang,
                max_tweets=self.max_tweets,
                batch_size=self.batch_size,
            )
            for date_ in daterange(self.since, self.until - timedelta(days=1))
        ]



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import datetime
from typing import Mapping, cast

from overrides import overrides
from typing_extensions import Final

from .._util.consts import TWITTER_CREATED_AT_FORMAT
from .._util.json_ import JsonSerializable
from .._util.typing_ import checked_cast

TweetId = str


class Tweet(JsonSerializable):
    """Data class to wrap Tweet JSON objects."""

    def __init__(self, json: Mapping[str, object]):
        self.json: Final = json

    @overrides
    def __repr__(self) -> str:
        obj = {
            "created_at": self.created_at,
            "id": self.id,
            "text": self.text,
            "user": self.user,
            "url": self.url,
        }
        return type(self).__name__ + repr(obj)

    @overrides
    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    @property
    def created_at(self) -> datetime:
        return datetime.strptime(
            checked_cast(str, self.json["created_at"]), TWITTER_CREATED_AT_FORMAT
        )

    @property
    def id(self) -> TweetId:
        return checked_cast(TweetId, self.json["id_str"])

    @property
    def text(self) -> str:
        return checked_cast(str, self.json["full_text"])

    @property
    def user(self) -> "User":
        return User(cast(Mapping[str, object], self.json["user"]))

    @property
    def url(self) -> str:
        return "{}/status/{}".format(self.user.url, self.id)

    @overrides
    def to_json(self) -> Mapping[str, object]:
        return self.json

    @classmethod
    @overrides
    def from_json(cls, obj: Mapping[str, object]) -> "Tweet":
        return cls(obj)


UserId = str


class User(JsonSerializable):
    """Data class to wrap Twitter user JSON objects."""

    def __init__(self, json: Mapping[str, object]):
        self.json: Final = json

    @overrides
    def __repr__(self) -> str:
        obj = {
            "id": self.id,
            "name": self.name,
            "screen_name": self.screen_name,
            "url": self.url,
        }
        return type(self).__name__ + repr(obj)

    @overrides
    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    @property
    def id(self) -> UserId:
        return checked_cast(UserId, self.json["id_str"])

    @property
    def name(self) -> str:
        return checked_cast(str, self.json["name"])

    @property
    def screen_name(self) -> str:
        return checked_cast(str, self.json["screen_name"])

    @property
    def url(self) -> str:
        return "https://twitter.com/{}".format(self.screen_name)

    @overrides
    def to_json(self) -> Mapping[str, object]:
        return self.json

    @classmethod
    @overrides
    def from_json(cls, obj: Mapping[str, object]) -> "User":
        return cls(obj)
